@@ -638,6 +638,47 @@ static void test_progress_callback(CuTest *tc)
     CuAssertTrue(tc, pb->read > 0);
 }
 
+/* Test that username:password components in url are ignored. */
+static void test_connection_userinfo_in_url(CuTest *tc)
+{
+    test_baton_t *tb;
+    apr_status_t status;
+    handler_baton_t handler_ctx[2];
+    const int num_requests = sizeof(handler_ctx)/sizeof(handler_ctx[0]);
+    int i;
+    progress_baton_t *pb;
+
+    test_server_message_t message_list[] = {
+        {CHUNKED_REQUEST(1, "1")},
+        {CHUNKED_REQUEST(1, "2")},
+    };
+
+    test_server_action_t action_list[] = {
+        {SERVER_RESPOND, CHUNKED_EMPTY_RESPONSE},
+        {SERVER_RESPOND, CHUNKED_RESPONSE(1, "2")},
+    };
+
+    apr_pool_t *test_pool = tc->testBaton;
+
+    /* Set up a test context with a server. */
+    status = test_http_server_setup(&tb,
+                                    message_list, num_requests,
+                                    action_list, num_requests, 0,
+                                    progress_conn_setup, test_pool);
+    CuAssertIntEquals(tc, APR_SUCCESS, status);
+
+    /* Create a connection using user:password@hostname syntax */
+    tb->serv_url = "http://user:password@localhost:" SERV_PORT_STR;
+    use_new_connection(tb, test_pool);
+
+    /* Send some requests on the connections */
+    for (i = 0 ; i < num_requests ; i++) {
+        create_new_request(tb, &handler_ctx[i], "GET", "/", i+1);
+    }
+
+    test_helper_run_requests_expect_ok(tc, tb, num_requests, handler_ctx,
+                                       test_pool);
+}
 
 /*****************************************************************************
  * Issue #91: test that serf correctly handle an incoming 4xx reponse while
@@ -1930,13 +1971,16 @@ ssltunnel_basic_authn_callback(char **username,
 /* Test if serf can successfully authenticate to a proxy used for an ssl
    tunnel. Retry the authentication a few times to test requeueing of the 
    CONNECT request. */
-static void ssltunnel_basic_auth(CuTest *tc, const char *resp_hdrs)
+static void ssltunnel_basic_auth(CuTest *tc, const char *server_resp_hdrs,
+                                 const char *proxy_resp_hdrs)
 {
     test_baton_t *tb;
     handler_baton_t handler_ctx[1];
     int num_requests_sent, num_requests_recvd;
     test_server_message_t message_list_server[2];
+    test_server_action_t action_list_proxy[7];
     test_server_action_t action_list_server[2];
+    apr_pool_t *test_pool = tc->testBaton;
     apr_status_t status;
 
     test_server_message_t message_list_proxy[] = {
@@ -1960,35 +2004,47 @@ static void ssltunnel_basic_auth(CuTest *tc, const char *resp_hdrs)
             "Proxy-Authorization: Basic c2VyZnByb3h5OnNlcmZ0ZXN0" CRLF
             CRLF },
     };
-    test_server_action_t action_list_proxy[] = {
-        {SERVER_RESPOND, "HTTP/1.1 407 Unauthorized" CRLF
-            "Transfer-Encoding: chunked" CRLF
-            "Proxy-Authenticate: Basic realm=""Test Suite Proxy""" CRLF
-            CRLF
-            "1" CRLF CRLF
-            "0" CRLF CRLF},
-        {SERVER_RESPOND, "HTTP/1.1 407 Unauthorized" CRLF
-            "Transfer-Encoding: chunked" CRLF
-            "Proxy-Authenticate: Basic realm=""Test Suite Proxy""" CRLF
-            CRLF
-            "1" CRLF CRLF
-            "0" CRLF CRLF},
-        {SERVER_RESPOND, "HTTP/1.1 407 Unauthorized" CRLF
-            "Transfer-Encoding: chunked" CRLF
-            "Proxy-Authenticate: Basic realm=""Test Suite Proxy""" CRLF
-            CRLF
-            "1" CRLF CRLF
-            "0" CRLF CRLF},
-        {SERVER_RESPOND, CHUNKED_EMPTY_RESPONSE},
-        /* Forward the remainder of the data to the server without validation */
-        {PROXY_FORWARD, "https://localhost:" SERV_PORT_STR},
-        /* If the client or the server closes the connection, stop forwarding.*/
-        {SERVER_RESPOND, CHUNKED_EMPTY_RESPONSE},
-        /* Again after disconnect. */
-        {PROXY_FORWARD, "https://localhost:" SERV_PORT_STR},
-    };
 
-    apr_pool_t *test_pool = tc->testBaton;
+    action_list_proxy[0].kind = SERVER_RESPOND;
+    action_list_proxy[0].text = apr_psprintf(test_pool,
+        "HTTP/1.1 407 Unauthorized" CRLF
+        "Transfer-Encoding: chunked" CRLF
+        "Proxy-Authenticate: Basic realm=""Test Suite Proxy""" CRLF
+        "%s"
+        CRLF
+        "1" CRLF CRLF
+        "0" CRLF CRLF, proxy_resp_hdrs);
+    action_list_proxy[1].kind = SERVER_RESPOND;
+    action_list_proxy[1].text = apr_psprintf(test_pool,
+        "HTTP/1.1 407 Unauthorized" CRLF
+        "Transfer-Encoding: chunked" CRLF
+        "Proxy-Authenticate: Basic realm=""Test Suite Proxy""" CRLF
+        "%s"
+        CRLF
+        "1" CRLF CRLF
+        "0" CRLF CRLF, proxy_resp_hdrs);
+
+    action_list_proxy[2].kind = SERVER_RESPOND;
+    action_list_proxy[2].text = apr_psprintf(test_pool,
+        "HTTP/1.1 407 Unauthorized" CRLF
+        "Transfer-Encoding: chunked" CRLF
+        "Proxy-Authenticate: Basic realm=""Test Suite Proxy""" CRLF
+        "%s"
+        CRLF
+        "1" CRLF CRLF
+        "0" CRLF CRLF, proxy_resp_hdrs);
+
+    action_list_proxy[3].kind = SERVER_RESPOND;
+    action_list_proxy[3].text = CHUNKED_EMPTY_RESPONSE;
+    /* Forward the remainder of the data to the server without validation */
+    action_list_proxy[4].kind = PROXY_FORWARD;
+    action_list_proxy[4].text = "https://localhost:" SERV_PORT_STR;
+    /* If the client or the server closes the connection, stop forwarding.*/
+    action_list_proxy[5].kind = SERVER_RESPOND;
+    action_list_proxy[5].text = CHUNKED_EMPTY_RESPONSE;
+    /* Again after disconnect. */
+    action_list_proxy[6].kind = PROXY_FORWARD;
+    action_list_proxy[6].text = "https://localhost:" SERV_PORT_STR;
 
     /* Make the server also require Basic authentication */
     message_list_server[0].text =
@@ -2019,7 +2075,7 @@ static void ssltunnel_basic_auth(CuTest *tc, const char *resp_hdrs)
         "%s"
         CRLF
         "1" CRLF CRLF
-        "0" CRLF CRLF, resp_hdrs);
+        "0" CRLF CRLF, server_resp_hdrs);
     action_list_server[1].kind = SERVER_RESPOND;
     action_list_server[1].text = CHUNKED_EMPTY_RESPONSE;
 
@@ -2064,12 +2120,20 @@ static void ssltunnel_basic_auth(CuTest *tc, const char *resp_hdrs)
 
 static void test_ssltunnel_basic_auth(CuTest *tc)
 {
-    ssltunnel_basic_auth(tc, "");
+    /* KeepAlive On for both proxy and server */
+    ssltunnel_basic_auth(tc, "", "");
 }
 
-static void test_ssltunnel_basic_auth_keepalive_off(CuTest *tc)
+static void test_ssltunnel_basic_auth_server_has_keepalive_off(CuTest *tc)
 {
-    ssltunnel_basic_auth(tc, "Connection: close" CRLF);
+    /* Add Connection:Close header to server response */
+    ssltunnel_basic_auth(tc, "Connection: close" CRLF, "");
+}
+
+static void test_ssltunnel_basic_auth_proxy_has_keepalive_off(CuTest *tc)
+{
+    /* Add Connection:Close header to proxy response */
+    ssltunnel_basic_auth(tc, "", "Connection: close" CRLF);
 }
 
 static apr_status_t
@@ -2138,12 +2202,14 @@ static void test_ssltunnel_digest_auth(CuTest *tc)
     /* Add a Basic header before Digest header, to test that serf uses the most
        secure authentication scheme first, instead of following the order of
        the headers. */
+    /* Use non standard case for Proxy-Authenticate header to test case
+       insensitivity for http headers. */
     test_server_action_t action_list_proxy[] = {
         {SERVER_RESPOND, "HTTP/1.1 407 Unauthorized" CRLF
             "Transfer-Encoding: chunked" CRLF
             "Proxy-Authenticate: Basic c2VyZjpzZXJmdGVzdA==" CRLF
             "Proxy-Authenticate: NonExistent blablablabla" CRLF
-            "Proxy-Authenticate: Digest realm=\"Test Suite Proxy\","
+            "proXy-Authenticate: Digest realm=\"Test Suite Proxy\","
             "nonce=\"ABCDEF1234567890\",opaque=\"myopaque\","
             "algorithm=\"MD5\",qop-options=\"auth\"" CRLF
             CRLF
@@ -2202,6 +2268,7 @@ CuSuite *test_context(void)
     SUITE_ADD_TEST(suite, test_request_timeout);
     SUITE_ADD_TEST(suite, test_connection_large_response);
     SUITE_ADD_TEST(suite, test_connection_large_request);
+    SUITE_ADD_TEST(suite, test_connection_userinfo_in_url);
     SUITE_ADD_TEST(suite, test_ssl_handshake);
     SUITE_ADD_TEST(suite, test_ssl_trust_rootca);
     SUITE_ADD_TEST(suite, test_ssl_application_rejects_cert);
@@ -2217,7 +2284,8 @@ CuSuite *test_context(void)
     SUITE_ADD_TEST(suite, test_setup_ssltunnel);
     SUITE_ADD_TEST(suite, test_ssltunnel_no_creds_cb);
     SUITE_ADD_TEST(suite, test_ssltunnel_basic_auth);
-    SUITE_ADD_TEST(suite, test_ssltunnel_basic_auth_keepalive_off);
+    SUITE_ADD_TEST(suite, test_ssltunnel_basic_auth_server_has_keepalive_off);
+    SUITE_ADD_TEST(suite, test_ssltunnel_basic_auth_proxy_has_keepalive_off);
     SUITE_ADD_TEST(suite, test_ssltunnel_digest_auth);
 
     return suite;
